# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
from typing import Tuple

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class LruActionCache(ActionCacheABC):
    """In-memory Action Cache implementation with LRU eviction.

    This cache has a configurable fixed size, evicting the least recently
    accessed entry when adding a new entry would exceed the fixed size. The
    cache is entirely stored in memory so its contents are lost on restart.

    This type of cache is ideal for use cases that need a simple and fast
    cache, with no requirements for longevity of the cache content. It is not
    recommended to use this type of cache in situations where you may wish to
    obtain cached results a reasonable time in the future, due to its fixed
    size.

    """

    def __init__(
        self, storage: StorageABC, max_cached_refs: int, allow_updates: bool = True, cache_failed_actions: bool = True
    ):
        """Initialise a new in-memory LRU Action Cache.

        Args:
            storage (StorageABC): Storage backend instance to be used.
            max_cached_refs (int): Maximum number of entries to store in the cache.
            allow_updates (bool): Whether to allow writing to the cache. If false,
                this is a read-only cache for all clients.
            cache_failed_actions (bool): Whether or not to cache Actions with
                non-zero exit codes.

        """
        super().__init__(storage=storage)

        self._cache_failed_actions = cache_failed_actions
        self._storage = storage
        self._allow_updates = allow_updates
        self._max_cached_refs = max_cached_refs
        self._digest_map = collections.OrderedDict()  # type: ignore

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached result for an Action.

        If there is no cached result found, returns None.

        Args:
            action_digest (Digest): The digest of the Action to retrieve the
                cached result of.

        """
        key = self._get_key(action_digest)
        if key in self._digest_map:
            assert self._storage, "Storage used before initialization"
            action_result = self._storage.get_message(self._digest_map[key], remote_execution_pb2.ActionResult)

            if action_result is not None:
                if self.referenced_blobs_still_exist(action_digest, action_result):
                    self._digest_map.move_to_end(key)
                    return action_result

                if self._allow_updates:
                    LOGGER.debug(
                        "Removing action digest from cache due to missing blobs in CAS.",
                        tags=dict(digest=action_digest),
                    )
                    del self._digest_map[key]

        raise NotFoundError(f"Key not found: {key}")

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores a result for an Action in the cache.

        If the result has a non-zero exit code and `cache_failed_actions` is False
        for this cache, the result is not cached.

        Args:
            action_digest (Digest): The digest of the Action whose result is
                being cached.
            action_result (ActionResult): The result to cache for the given
                Action digest.

        """
        if self._cache_failed_actions or action_result.exit_code == 0:
            key = self._get_key(action_digest)
            if not self._allow_updates:
                raise NotImplementedError("Updating cache not allowed")

            if self._max_cached_refs == 0:
                return

            while len(self._digest_map) >= self._max_cached_refs:
                self._digest_map.popitem(last=False)

            assert self._storage, "Storage used before initialization"
            result_digest = self._storage.put_message(action_result)
            self._digest_map[key] = result_digest

            LOGGER.info("Result cached for action.", tags=dict(digest=action_digest))

    def _get_key(self, action_digest: Digest) -> Tuple[str, int]:
        """Get a hashable cache key from a given Action digest.

        Args:
            action_digest (Digest): The digest to produce a cache key for.

        """
        return (action_digest.hash, action_digest.size_bytes)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Log Stream API"""
import abc
import buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2
import grpc.aio as grpc
import typing

class LogStreamServiceStub:
    """#### Introduction

    The Log Stream API manages LogStream resources which are used to stream
    writes and reads of an ordered sequence of bytes of unknown eventual length.

    Note that this is an API Interface and not an API Service, per the definitons
    at: https://cloud.google.com/apis/design/glossary

    Log Stream API supports the reading of unfinalized LogStreams either by
    seeking or in "tail" mode, for example by end-users browsing to a build
    result UI interested in seeing logs from a build action as soon as they are
    (or as they become) available.

    Reads and Writes of LogStreams are done via the Byte Stream API:
    https://cloud.google.com/dataproc/docs/reference/rpc/google.bytestream
    https://github.com/googleapis/googleapis/blob/master/google/bytestream/bytestream.proto

    #### Writing LogStreams

    LogStreams are written to via the Byte Stream API's `Write` RPC. Bytes
    written to LogStreams are expected to be committed and available for reading
    within a reasonable period of time (implementation-defined). Committed bytes
    to a LogStream cannot be overwritten, and finalized LogStreams - indicated by
    setting `finish_write` field in the final WriteRequest - also cannot be
    appended to.

    When calling the Byte Stream API's `Write` RPC to write LogStreams, writers
    must pass the `write_resource_name` of a LogStream as
    `ByteStream.WriteRequest.resource_name` rather than the LogStream's `name`.
    Separate resource names for reading and writing allows for broadcasting the
    read resource name widely while simultaneously ensuring that only writer(s)
    with knowledge of the write resource name may have written bytes to the
    LogStream.

    #### Reading LogStreams

    Use the Byte Stream API's `Read` RPC to read LogStreams. When reading
    finalized LogStreams the server will stream all contents of the LogStream
    starting at `ByteStream.ReadRequest.read_offset`.

    When reading unfinalized LogStreams the server must keep the streaming
    `ByteStream.Read` RPC open and send `ByteStream.ReadResponse` messages as
    more bytes become available or the LogStream is finalized.

    #### Example Multi-Party Read/Write Flow

    1. LogStream Writer calls `CreateLogStream`
    2. LogStream Writer publishes `LogStream.name`
    3. LogStream Writer calls `ByteStream.Write` with
       `LogStream.write_resource_name` as
       `ByteStream.WriteRequest.resource_name`,
       `ByteStream.WriteRequest.finish_write`=false.
    4. LogStream Reader(s) call `ByteStream.Read` with the published
       `LogStream.name` as `ByteStream.ReadRequest.resource_name`.
    5. LogStream Service streams all committed bytes to LogStream Reader(s),
       leave the stream open.
    6. LogStream Writer calls `ByteStream.Write` with
       `LogStream.write_resource_name` as
       `ByteStream.WriteRequest.resource_name`,
       `ByteStream.WriteRequest.finish_write`=true.
    7. LogStream Service streams all remaining bytes to LogStream Reader(s),
       terminates the stream.
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    CreateLogStream: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2.CreateLogStreamRequest,
        buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2.LogStream,
    ]
    """Create a LogStream which may be written to.

    The returned LogStream resource name will include a `write_resource_name`
    which is the resource to use when writing to the LogStream.
    Callers of CreateLogStream are expected to NOT publish the
    `write_resource_name`.
    """

class LogStreamServiceServicer(metaclass=abc.ABCMeta):
    """#### Introduction

    The Log Stream API manages LogStream resources which are used to stream
    writes and reads of an ordered sequence of bytes of unknown eventual length.

    Note that this is an API Interface and not an API Service, per the definitons
    at: https://cloud.google.com/apis/design/glossary

    Log Stream API supports the reading of unfinalized LogStreams either by
    seeking or in "tail" mode, for example by end-users browsing to a build
    result UI interested in seeing logs from a build action as soon as they are
    (or as they become) available.

    Reads and Writes of LogStreams are done via the Byte Stream API:
    https://cloud.google.com/dataproc/docs/reference/rpc/google.bytestream
    https://github.com/googleapis/googleapis/blob/master/google/bytestream/bytestream.proto

    #### Writing LogStreams

    LogStreams are written to via the Byte Stream API's `Write` RPC. Bytes
    written to LogStreams are expected to be committed and available for reading
    within a reasonable period of time (implementation-defined). Committed bytes
    to a LogStream cannot be overwritten, and finalized LogStreams - indicated by
    setting `finish_write` field in the final WriteRequest - also cannot be
    appended to.

    When calling the Byte Stream API's `Write` RPC to write LogStreams, writers
    must pass the `write_resource_name` of a LogStream as
    `ByteStream.WriteRequest.resource_name` rather than the LogStream's `name`.
    Separate resource names for reading and writing allows for broadcasting the
    read resource name widely while simultaneously ensuring that only writer(s)
    with knowledge of the write resource name may have written bytes to the
    LogStream.

    #### Reading LogStreams

    Use the Byte Stream API's `Read` RPC to read LogStreams. When reading
    finalized LogStreams the server will stream all contents of the LogStream
    starting at `ByteStream.ReadRequest.read_offset`.

    When reading unfinalized LogStreams the server must keep the streaming
    `ByteStream.Read` RPC open and send `ByteStream.ReadResponse` messages as
    more bytes become available or the LogStream is finalized.

    #### Example Multi-Party Read/Write Flow

    1. LogStream Writer calls `CreateLogStream`
    2. LogStream Writer publishes `LogStream.name`
    3. LogStream Writer calls `ByteStream.Write` with
       `LogStream.write_resource_name` as
       `ByteStream.WriteRequest.resource_name`,
       `ByteStream.WriteRequest.finish_write`=false.
    4. LogStream Reader(s) call `ByteStream.Read` with the published
       `LogStream.name` as `ByteStream.ReadRequest.resource_name`.
    5. LogStream Service streams all committed bytes to LogStream Reader(s),
       leave the stream open.
    6. LogStream Writer calls `ByteStream.Write` with
       `LogStream.write_resource_name` as
       `ByteStream.WriteRequest.resource_name`,
       `ByteStream.WriteRequest.finish_write`=true.
    7. LogStream Service streams all remaining bytes to LogStream Reader(s),
       terminates the stream.
    """

    @abc.abstractmethod
    async def CreateLogStream(
        self,
        request: buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2.CreateLogStreamRequest,
        context: grpc.ServicerContext[typing.Any, typing.Any],
    ) -> buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2.LogStream:
        """Create a LogStream which may be written to.

        The returned LogStream resource name will include a `write_resource_name`
        which is the resource to use when writing to the LogStream.
        Callers of CreateLogStream are expected to NOT publish the
        `write_resource_name`.
        """

def add_LogStreamServiceServicer_to_server(servicer: LogStreamServiceServicer, server: grpc.Server) -> None: ...



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import contextmanager
from typing import Iterator, Optional

import grpc
from grpc import RpcError

from buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2 import CreateLogStreamRequest, LogStream
from buildgrid._protos.build.bazel.remote.logstream.v1.remote_logstream_pb2_grpc import LogStreamServiceStub
from buildgrid.server.app.commands.cmd_logstream import instanced_resource_name
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class LogStreamClient:
    def __init__(self, channel: grpc.Channel, instance_name: str = "") -> None:
        self._channel = channel
        self._instance_name = instance_name
        self._logstream_stub: Optional[LogStreamServiceStub] = LogStreamServiceStub(self._channel)

    def create(self, parent: str) -> LogStream:
        assert self._logstream_stub, "LogStreamClient used after close"

        parent = instanced_resource_name(self._instance_name, parent)
        request = CreateLogStreamRequest(parent=parent)
        try:
            return self._logstream_stub.CreateLogStream(request)
        except RpcError as e:
            LOGGER.exception(f"Error creating a LogStream: {e.details()}")
            raise ConnectionError(e.details())

    def close(self) -> None:
        self._logstream_stub = None


@contextmanager
def logstream_client(channel: grpc.Channel, instance_name: str) -> Iterator[LogStreamClient]:
    client = LogStreamClient(channel, instance_name)
    try:
        yield client
    finally:
        client.close()



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import Any, Optional, Set, TypeVar

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Action, ActionResult, Digest, Tree
from buildgrid.server.cas.instance import EMPTY_BLOB_DIGEST
from buildgrid.server.cas.storage.storage_abc import StorageABC
from buildgrid.server.exceptions import NotFoundError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric
from buildgrid.server.utils.digests import HashableDigest

LOGGER = buildgrid_logger(__name__)


T = TypeVar("T", bound="ActionCacheABC")


class ActionCacheABC(ABC):
    def __init__(self, allow_updates: bool = False, storage: Optional[StorageABC] = None):
        self._allow_updates = allow_updates
        self._storage = storage

    @property
    def allow_updates(self) -> bool:
        return self._allow_updates

    def __enter__(self: T) -> T:
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        if self._storage is not None:
            self._storage.start()

    def stop(self) -> None:
        if self._storage is not None:
            self._storage.stop()
        LOGGER.info("Stopped ActionCache.")

    @abstractmethod
    def get_action_result(self, action_digest: Digest) -> ActionResult:
        raise NotImplementedError()

    @abstractmethod
    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        raise NotImplementedError()

    def referenced_blobs_still_exist(self, action_digest: Digest, action_result: ActionResult) -> bool:
        if self._referenced_blobs_still_exist(action_digest, action_result):
            return True

        publish_counter_metric(METRIC.ACTION_CACHE.INVALID_CACHE_COUNT, 1)
        return False

    def _referenced_blobs_still_exist(self, action_digest: Digest, action_result: ActionResult) -> bool:
        """Checks CAS for Action and ActionResult output blobs existence.

        Args:
            action_digest (Digest): Digest for the Action whose top level digests
            will be searched for.
            action_result (ActionResult): ActionResult to search referenced
            output blobs for.

        Returns:
            True if all referenced blobs are present in CAS, False otherwise.
        """
        if not self._storage:
            return True
        blobs_needed: Set[HashableDigest] = set()

        for output_file in action_result.output_files:
            blobs_needed.add(HashableDigest(output_file.digest.hash, output_file.digest.size_bytes))

        for output_directory in action_result.output_directories:
            if output_directory.HasField("tree_digest"):
                blobs_needed.add(
                    HashableDigest(output_directory.tree_digest.hash, output_directory.tree_digest.size_bytes)
                )
                tree = self._storage.get_message(output_directory.tree_digest, Tree)
                if tree is None:
                    return False

                for file_node in tree.root.files:
                    blobs_needed.add(HashableDigest(file_node.digest.hash, file_node.digest.size_bytes))

                for child in tree.children:
                    for file_node in child.files:
                        blobs_needed.add(HashableDigest(file_node.digest.hash, file_node.digest.size_bytes))
            elif output_directory.HasField("root_directory_digest"):
                blobs_needed.add(
                    HashableDigest(
                        output_directory.root_directory_digest.hash, output_directory.root_directory_digest.size_bytes
                    )
                )
                try:
                    for directory in self._storage.get_tree(
                        output_directory.root_directory_digest, raise_on_missing_subdir=True
                    ):
                        blobs_needed.update(
                            [
                                HashableDigest(file_node.digest.hash, file_node.digest.size_bytes)
                                for file_node in directory.files
                            ]
                        )
                        blobs_needed.update(
                            [
                                HashableDigest(dir_node.digest.hash, dir_node.digest.size_bytes)
                                for dir_node in directory.directories
                            ]
                        )
                except NotFoundError:
                    return False

        if action_result.stdout_digest.hash and not action_result.stdout_raw:
            blobs_needed.add(HashableDigest(action_result.stdout_digest.hash, action_result.stdout_digest.size_bytes))

        if action_result.stderr_digest.hash and not action_result.stderr_raw:
            blobs_needed.add(HashableDigest(action_result.stderr_digest.hash, action_result.stderr_digest.size_bytes))

        # Additionally refresh the TTL of the ActionDigest and the top level digests
        # contained within. This will keep the Action around for use cases like bgd-browser
        # where you want to look at both the Action and it's ActionResult, but with minimal
        # overhead.
        action = self._storage.get_message(action_digest, Action)
        action_blobs: Set[HashableDigest] = set()
        if action:
            action_blobs.add(HashableDigest(action_digest.hash, action_digest.size_bytes))
            action_blobs.add(HashableDigest(action.command_digest.hash, action.command_digest.size_bytes))
            action_blobs.add(HashableDigest(action.input_root_digest.hash, action.input_root_digest.size_bytes))

        blobs_to_check: Set[HashableDigest] = blobs_needed | action_blobs
        # No need to check the underlying storage for the empty blob as it is a special case blob which always exists
        # It is possible that the empty blob is not actually present in the underlying storage
        blobs_to_check.discard(HashableDigest(EMPTY_BLOB_DIGEST.hash, EMPTY_BLOB_DIGEST.size_bytes))
        missing = self._storage.missing_blobs([blob.to_digest() for blob in blobs_to_check])
        required_missing = [blob for blob in missing if HashableDigest(blob.hash, blob.size_bytes) in blobs_needed]
        if len(required_missing) != 0:
            LOGGER.debug(
                "Missing blobs.", tags=dict(required_missing=len(required_missing), blobs_needed=len(blobs_needed))
            )
            return False
        return True



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Any, Callable, Iterator, Optional, TypeVar, cast

import grpc

from buildgrid.server.context import current_instance, current_method, current_service, method_context, service_context
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metadata import extract_client_identity_dict, extract_request_metadata_dict
from buildgrid.server.metrics_names import METRIC

from .authorize import authorize
from .errors import handle_errors
from .instance import instanced
from .io import network_io
from .metadata import metadatacontext
from .requestid import track_request_id
from .time import timed

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]
LOGGER = buildgrid_logger(__name__)


def log_rpc(f: Func) -> Func:
    @functools.wraps(f)
    def server_stream_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Iterator[Any]:
        tags = dict(
            service=current_service(),
            method=current_method(),
            peer=context.peer(),
            **extract_request_metadata_dict(context.invocation_metadata()),
            **extract_client_identity_dict(current_instance(), context.invocation_metadata()),
        )
        LOGGER.info("Received request.", tags=tags)
        yield from f(self, message, context)

    @functools.wraps(f)
    def server_unary_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Any:
        tags = dict(
            service=current_service(),
            method=current_method(),
            peer=context.peer(),
            **extract_request_metadata_dict(context.invocation_metadata()),
            **extract_client_identity_dict(current_instance(), context.invocation_metadata()),
        )
        LOGGER.info("Received request.", tags=tags)
        return f(self, message, context)

    if inspect.isgeneratorfunction(f):
        return cast(Func, server_stream_wrapper)
    return cast(Func, server_unary_wrapper)


def named_rpc(f: Func) -> Func:
    @functools.wraps(f)
    def server_stream_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Iterator[Any]:
        with service_context(self.SERVICE_NAME), method_context(f.__name__):
            yield from f(self, message, context)

    @functools.wraps(f)
    def server_unary_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Any:
        with service_context(self.SERVICE_NAME), method_context(f.__name__):
            return f(self, message, context)

    if inspect.isgeneratorfunction(f):
        return cast(Func, server_stream_wrapper)
    return cast(Func, server_unary_wrapper)


def rpc(
    *,
    instance_getter: Optional[Callable[[Any], str]] = None,
    request_formatter: Callable[[Any], Any] = lambda r: str(r),
) -> Callable[[Func], Func]:
    """
    The RPC decorator provides common functionality to all buildgrid servicer methods.
    This decorator should be attached to all endpoints in the application.

    All endpoints will produce the following metrics, with a tag "code" for grpc status values,
    a tag "service" for the service name of the RPC, and a tag "method" for the RPC method name.
    * ``rpc.duration.ms``: The time in milliseconds spent on the method.
    * ``rpc.input_bytes.count``: The number of message bytes sent from client to server.
    * ``rpc.output_bytes.count``: The number of message bytes sent from server to client.

    All other metrics produced during an RPC will have the tags "service" and "method" attached.

    Args:
        instance_getter (Callable[[Any], str]): Determines how to fetch the instance name
            from the request payload. If provided, the tag "instance" will also be applied for all
            metrics, and the RPC will be enrolled in authentication/authorization.

        request_formatter (Callable[[Any], str]): Determines how to format the request payloads in logs.
    """

    def decorator(func: Func) -> Func:
        # Note: decorators are applied in reverse order from a normal decorator pattern.
        #   All decorators that apply context vars are invoked earlier in the call chain
        #   such that the context vars remain available for logging and metrics population.
        func = log_rpc(func)
        if instance_getter:
            func = authorize(func)
        func = handle_errors(request_formatter)(func)
        func = network_io(func)
        func = timed(METRIC.RPC.DURATION)(func)
        func = metadatacontext(func)
        func = track_request_id(func)
        func = named_rpc(func)
        if instance_getter:
            func = instanced(instance_getter)(func)

        return func

    return decorator



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Any, Callable, Iterator, TypeVar, cast

import grpc

from buildgrid.server.metadata import ctx_request_metadata, extract_request_metadata

Func = TypeVar("Func", bound=Callable)  # type: ignore[type-arg]


def metadatacontext(f: Func) -> Func:
    """Helper function to obtain metadata and set request metadata ContextVar,
    and then reset it on completion of method.

    Note:
        args[2] of the method must be of type grpc.ServicerContext

    This returns a decorator that extracts the invocation_metadata from the
    context argument and sets the ContextVar variable with it. Resetting the
    ContextVar variable after the method has completed.
    """

    @functools.wraps(f)
    def server_stream_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Iterator[Any]:
        metadata = extract_request_metadata(context.invocation_metadata())
        token = ctx_request_metadata.set(metadata)
        try:
            yield from f(self, message, context)
        finally:
            ctx_request_metadata.reset(token)

    @functools.wraps(f)
    def server_unary_wrapper(self: Any, message: Any, context: grpc.ServicerContext) -> Any:
        metadata = extract_request_metadata(context.invocation_metadata())
        token = ctx_request_metadata.set(metadata)
        try:
            return f(self, message, context)
        finally:
            ctx_request_metadata.reset(token)

    if inspect.isgeneratorfunction(f):
        return cast(Func, server_stream_wrapper)
    return cast(Func, server_unary_wrapper)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2021 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import buildgrid._protos.google.devtools.build.v1.build_events_pb2
import buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class QueryEventStreamsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUILD_ID_PATTERN_FIELD_NUMBER: builtins.int
    build_id_pattern: builtins.str
    def __init__(
        self,
        *,
        build_id_pattern: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["build_id_pattern", b"build_id_pattern"]) -> None: ...

global___QueryEventStreamsRequest = QueryEventStreamsRequest

@typing_extensions.final
class QueryEventStreamsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class BuildEventStream(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STREAM_ID_FIELD_NUMBER: builtins.int
        EVENTS_FIELD_NUMBER: builtins.int
        @property
        def stream_id(self) -> buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId: ...
        @property
        def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2.OrderedBuildEvent]: ...
        def __init__(
            self,
            *,
            stream_id: buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId | None = ...,
            events: collections.abc.Iterable[buildgrid._protos.google.devtools.build.v1.publish_build_event_pb2.OrderedBuildEvent] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["stream_id", b"stream_id"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["events", b"events", "stream_id", b"stream_id"]) -> None: ...

    STREAMS_FIELD_NUMBER: builtins.int
    @property
    def streams(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___QueryEventStreamsResponse.BuildEventStream]: ...
    def __init__(
        self,
        *,
        streams: collections.abc.Iterable[global___QueryEventStreamsResponse.BuildEventStream] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["streams", b"streams"]) -> None: ...

global___QueryEventStreamsResponse = QueryEventStreamsResponse



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
DiskStorage
==================

A CAS storage provider that stores files as blobs on disk.
"""

import errno
import io
import os
import tempfile
from typing import IO, Dict, List, Optional, Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.exceptions import StorageFullError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.settings import HASH, MAX_IN_MEMORY_BLOB_SIZE_BYTES

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)


class DiskStorage(StorageABC):
    TYPE = "Disk"

    def __init__(self, path: str) -> None:
        if not os.path.isabs(path):
            self.__root_path = os.path.abspath(path)
        else:
            self.__root_path = path
        self.__cas_path = os.path.join(self.__root_path, "cas")

        self.objects_path = os.path.join(self.__cas_path, "objects")
        self.temp_path = os.path.join(self.__root_path, "tmp")

        os.makedirs(self.objects_path, exist_ok=True)
        os.makedirs(self.temp_path, exist_ok=True)

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        LOGGER.debug("Checking for blob.", tags=dict(digest=digest))
        return os.path.exists(self._get_object_path(digest))

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        return [digest for digest in digests if not os.path.exists(self._get_object_path(digest))]

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        LOGGER.debug("Getting blob.", tags=dict(digest=digest))
        try:
            f = open(self._get_object_path(digest), "rb")
            # TODO probably need to make StorageABC generic...?
            return io.BufferedReader(f)  # type: ignore[arg-type]
        except FileNotFoundError:
            return None

    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        result = {}
        for digest in digests:
            try:
                with open(self._get_object_path(digest), "rb") as f:
                    result[digest.hash] = f.read()
            except FileNotFoundError:
                # Ignore files not found, will be reported as NOT_FOUND higher up.
                pass
        return result

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        LOGGER.debug("Deleting blob.", tags=dict(digest=digest))
        try:
            os.remove(self._get_object_path(digest))
        except OSError:
            pass

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        failed_deletions = []
        for digest in digests:
            try:
                os.remove(self._get_object_path(digest))
            except FileNotFoundError:
                # Ignore files not found. Already deleted.
                pass
            except OSError:
                # If deletion threw an exception, assume deletion failed. More specific implementations
                # with more information can return if a blob was missing instead
                LOGGER.warning("Unable to clean up digest.", tags=dict(digest=digest), exc_info=True)
                failed_deletions.append(digest.hash)

        return failed_deletions

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        LOGGER.debug("Writing blob.", tags=dict(digest=digest))
        object_path = self._get_object_path(digest)

        write_session.seek(0)
        try:
            with tempfile.NamedTemporaryFile("wb", dir=self.temp_path) as f:
                while data := write_session.read(MAX_IN_MEMORY_BLOB_SIZE_BYTES):
                    f.write(data)
                os.makedirs(os.path.dirname(object_path), exist_ok=True)
                os.link(f.name, object_path)
        except FileExistsError:
            # Object is already there!
            pass
        except OSError as e:
            # Not enough space error or file too large
            if e.errno in [errno.ENOSPC, errno.EFBIG]:
                raise StorageFullError(f"Disk Error: {e.errno}") from e
            raise e

    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        result = []
        for digest, data in blobs:
            object_path = self._get_object_path(digest)
            if len(data) != digest.size_bytes or HASH(data).hexdigest() != digest.hash:
                result.append(Status(code=code_pb2.INVALID_ARGUMENT, message="Data doesn't match hash"))
                continue
            try:
                with tempfile.NamedTemporaryFile("wb", dir=self.temp_path) as f:
                    f.write(data)
                    os.makedirs(os.path.dirname(object_path), exist_ok=True)
                    os.link(f.name, object_path)
                result.append(Status(code=code_pb2.OK))
            except FileExistsError:
                # Object is already there!
                result.append(Status(code=code_pb2.OK))
            except OSError as e:
                code = code_pb2.INTERNAL
                if e.errno in [errno.ENOSPC, errno.EFBIG]:
                    code = code_pb2.RESOURCE_EXHAUSTED
                result.append(Status(code=code, message=f"Disk Error: {e}"))
        return result

    def _get_object_path(self, digest: Digest) -> str:
        return os.path.join(self.objects_path, digest.hash[:2], digest.hash[2:])
